#include "phero_entity.h"
#include <argos3/core/simulator/simulator.h>
#include <argos3/core/simulator/space/space.h>
#include <argos3/core/simulator/loop_functions.h>


namespace argos {
 
  /****************************************/
  /****************************************/

  class CPheroFromLoopFunctions : public CPheroEntity::CPheroSource {
 
  public:
 
  CPheroFromLoopFunctions(UInt32 un_pixels_per_meter) :
  m_cLoopFunctions(CSimulator::GetInstance().GetLoopFunctions()),
  m_unPixelsPerMeter(un_pixels_per_meter) {
  const CVector3& cArenaSize = CSimulator::GetInstance().GetSpace().GetArenaSize();
  m_cHalfArenaSize.Set(
  cArenaSize.GetX() * 0.5f,
  cArenaSize.GetY() * 0.5f);
  const CVector3& cArenaCenter = CSimulator::GetInstance().GetSpace().GetArenaCenter();
  m_cArenaCenter.Set(cArenaCenter.GetX(),
  cArenaCenter.GetY());
  }
  // Getting Pheromone at the point (x, y) in the arena
  virtual Real GetPheroAtPoint(Real f_x,
  Real f_y) {
  return m_cLoopFunctions.GetPhero(CVector2(f_x, f_y)); // 20200525 I need to define a method retrieving pheromone
  }
 
  private:
 
  CLoopFunctions& m_cLoopFunctions;
  UInt32 m_unPixelsPerMeter;
  CVector2 m_cHalfArenaSize;
  CVector2 m_cArenaCenter;
  };
 
  /****************************************/
  /****************************************/
 
  CFloorEntity::CFloorEntity() :
  CEntity(NULL),
  m_eColorSource(UNSET),
  m_pcColorSource(NULL),
  m_bHasChanged(true) {}
 
  /****************************************/
  /****************************************/
 
  CPheroEntity::CPheroEntity(const std::string& str_id,
  UInt32 un_pixels_per_meter) :
  CEntity(NULL, str_id),
  m_eColorSource(FROM_LOOP_FUNCTIONS),
  m_pcColorSource(new CFloorColorFromLoopFunctions(un_pixels_per_meter)),
  m_bHasChanged(true) {}
 
  /****************************************/
  /****************************************/
 
  CPheroEntity::~CPheroEntity() {
  if(m_pcColorSource != NULL) {
  delete m_pcColorSource;
  }
  }
 
  /****************************************/
  /****************************************/
 
  void CPheroEntity::Init(TConfigurationNode& t_tree) {
  /* Init parent */
  CEntity::Init(t_tree);
  /* Parse XML */
  std::string strColorSource;
  GetNodeAttribute(t_tree, "source", strColorSource);
  if(strColorSource == "loop_functions") {
  m_eColorSource = FROM_LOOP_FUNCTIONS;
  UInt32 unPixelsPerMeter;
  GetNodeAttribute(t_tree, "pixels_per_meter", unPixelsPerMeter);
  m_pcColorSource = new CFloorColorFromLoopFunctions(unPixelsPerMeter);
  }
  else if(strColorSource == "image") {
 #ifdef ARGOS_WITH_FREEIMAGE
  m_eColorSource = FROM_IMAGE;
  std::string strPath;
  GetNodeAttribute(t_tree, "path", strPath);
  ExpandEnvVariables(strPath);
  m_pcColorSource = new CFloorColorFromImageFile(strPath);
 #else
  THROW_ARGOSEXCEPTION("ARGoS was compiled without FreeImage, this image source is unsupported for the floor entity \"" <<
  GetId() <<
  "\"");
 #endif
  }
  else {
  THROW_ARGOSEXCEPTION("Unknown image source \"" <<
  strColorSource <<
  "\" for the floor entity \"" <<
  GetId() <<
  "\"");
  }
  }
 
  /****************************************/
  /****************************************/
 
  void CPheroEntity::Reset() {
  m_pcColorSource->Reset();
  }
 
  /****************************************/
  /****************************************/
 
 #ifdef ARGOS_WITH_FREEIMAGE
  void CPheroEntity::SaveAsImage(const std::string& str_path) {
  m_pcColorSource->SaveAsImage(str_path);
  }
 #endif
 
  /****************************************/
  /****************************************/
 
  REGISTER_ENTITY(CFloorEntity,
  "floor",
  "Carlo Pinciroli [ilpincy@gmail.com]",
  "1.0",
  "It contains the properties of the arena floor.",
  "The floor entity contains the properties of the arena floor. In the current\n"
  "implementation, it contains only the color of the floor. The floor color is\n"
  "detected by the robots' ground sensors.\n\n"
  "REQUIRED XML CONFIGURATION\n\n"
  " <arena ...>\n"
  " ...\n"
  " <floor id=\"floor\"\n"
  " source=\"SOURCE\" />\n"
  " ...\n"
  " </arena>\n\n"
  "The 'id' attribute is necessary and must be unique among the entities. If two\n"
  "entities share the same id, initialization aborts.\n"
  "The 'source' attribute specifies where to get the color of the floor from. Its\n"
  "value, here denoted as SOURCE, can assume the following values:\n\n"
  " image The color is calculated from the passed image file\n"
  " loop_functions The color is calculated calling the loop functions\n\n"
  "When 'source' is set to 'image', as showed in the following example, you have\n"
  "to specify the image path in the additional attribute 'path':\n\n"
  " <arena ...>\n"
  " ...\n"
  " <floor id=\"floor\"\n"
  " source=\"image\"\n"
  " path=\"/path/to/imagefile.ext\" />\n"
  " ...\n"
  " </arena>\n\n"
  "Many image formats are available, such as PNG, JPG, BMP, GIF and many more.\n"
  "Refer to the FreeImage webpage for a complete list of supported image formats\n"
  "(http://freeimage.sourceforge.net/features.html).\n\n"
  "When 'source' is set to 'loop_functions', as showed in the following example,\n"
  "an image is implicitly created to be used as texture for graphical\n"
  "visualizations. The algorithm that creates the texture needs to convert from\n"
  "meters (in the arena) to pixels (of the texture). You control how many pixels\n"
  "per meter are used with the attribute 'pixels_per_meter'. Clearly, the higher\n"
  "value, the higher the quality, but also the slower the algorithm and the bigger\n"
  "the texture. The algorithm is called only once at init time, so the fact that\n"
  "it is slow is not so important. However, the image size is limited by OpenGL.\n"
  "Every implementation has its own limit, and you should check yours if any\n"
  "texture-related problem arises. Now for the example:\n\n"
  " <arena ...>\n"
  " ...\n"
  " <floor id=\"floor\"\n"
  " source=\"loop_functions\"\n"
  " pixels_per_meter=\"100\" />\n"
  " ...\n"
  " </arena>\n\n"
  "OPTIONAL XML CONFIGURATION\n\n"
  "None for the time being.\n",
  "Usable"
  );
 
  /****************************************/
  /****************************************/
 
  class CSpaceOperationAddCFloorEntity : public CSpaceOperationAddEntity {
  public:
  void ApplyTo(CSpace& c_space, CFloorEntity& c_entity) {
  c_space.AddEntity(c_entity);
  c_space.SetFloorEntity(c_entity);
  }
  };
 
  REGISTER_SPACE_OPERATION(CSpaceOperationAddEntity,
  CSpaceOperationAddCFloorEntity,
  CFloorEntity);
  REGISTER_STANDARD_SPACE_OPERATION_REMOVE_ENTITY(CFloorEntity);
 
  /****************************************/
  /****************************************/
 
 }